/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.weaver;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import jp.objectfanatics.assertion.weaver.Weaver.NoWeavingNeededException;
import jp.objectfanatics.assertion.weaver.Weaver.NotClassFileException;

import org.apache.commons.io.IOUtils;

/**
 * {@link Weaver} for class file.
 * 
 * @author Makoto Sato
 */
public class ClassFileWeaver {
	
	/**
	 * A {@link Weaver} configured with classpaths.
	 */
	private final Weaver weaver;
	
	/**
	 * Constructor.
	 * 
	 * @param classpaths
	 *            classpaths to be used to create {@link Weaver} instance.
	 */
	public ClassFileWeaver(String... classpaths) {
		this.weaver = new Weaver(classpaths);
	}
	
	/**
	 * Weaves assertion functions into a class file if needed.
	 * 
	 * @param file
	 *            input and output file
	 * @throws NoWeavingNeededException
	 * @throws IOException
	 */
	public void weave(File file) throws NoWeavingNeededException, IOException, NotClassFileException {
		weave(file, file);
	}
	
	/**
	 * Weaves assertion functions into a class file if needed.
	 * 
	 * @param inFile
	 *            input class file
	 * @param outFile
	 *            output class file
	 * @throws NoWeavingNeededException
	 * @throws IOException
	 */
	public void weave(File inFile, File outFile) throws NoWeavingNeededException, IOException, NotClassFileException {
		byte[] originalBytes = readBytes(inFile);
		byte[] wovenBytes = weaver.weave(originalBytes);
		writeBytes(outFile, wovenBytes);
	}
	
	/**
	 * Returns contents of a file as a byte[].
	 * 
	 * @param file
	 * @return contents of a file as a byte[].
	 * @throws IOException
	 */
	private byte[] readBytes(File file) throws IOException {
		InputStream is = null;
		try {
			is = new BufferedInputStream(new FileInputStream(file));
			return IOUtils.toByteArray(is);
		} finally {
			if (is != null) is.close();
		}
	}
	
	/**
	 * Writes bytes from bytes to file.
	 * 
	 * @param file
	 * @param bytes
	 * @throws IOException
	 */
	private void writeBytes(File file, byte[] bytes) throws IOException {
		OutputStream os = null;
		try {
			os = new FileOutputStream(file);
			IOUtils.write(bytes, os);
		} finally {
			if (os != null) os.close();
		}
	}
}
