/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.weaver;

import java.io.File;
import java.io.IOException;

import jp.objectfanatics.assertion.weaver.Weaver.IllegalParameterOverridingException;
import jp.objectfanatics.assertion.weaver.Weaver.IllegalReturnValueOverridingException;
import jp.objectfanatics.assertion.weaver.Weaver.IllegalUseOfConstraintAnnotationException;
import jp.objectfanatics.assertion.weaver.Weaver.NoSuitableAnnotatedParameterTypeException;
import jp.objectfanatics.assertion.weaver.Weaver.NoSuitableAnnotatedReturnTypeException;
import jp.objectfanatics.assertion.weaver.Weaver.NoWeavingNeededException;
import jp.objectfanatics.assertion.weaver.Weaver.NotClassFileException;

/**
 * {@link Weaver} for class folder.
 * 
 * @author Makoto Sato
 */
public class ClassFolderWeaver {
	
	/**
	 * A {@link Weaver} configured with classpaths.
	 */
	private final ClassFileWeaver classFileWeaver;
	
	/**
	 * Whether there are any illegal annotation definitions.
	 */
	private boolean hasIllegalConstraintAnnotationDefinition = false;
	
	private final WeavingStatusListener listener;
	
	/**
	 * Constructor.
	 * 
	 * @param classpaths
	 *            classpaths to be used to create {@link Weaver} instance.
	 */
	public ClassFolderWeaver(WeavingStatusListener listener, String... classpaths) {
		this.listener = listener;
		this.classFileWeaver = new ClassFileWeaver(classpaths);
	}
	
	/**
	 * Weaves assertion functions into a class or class files in a class folder if needed.
	 * 
	 * @param classFolderOrClassFilePath
	 *            a path of a class folder or a class file
	 * @throws IOException
	 */
	public void weave(String classFolderOrClassFilePath) throws IOException {
		weave(new File(classFolderOrClassFilePath));
		if (hasIllegalConstraintAnnotationDefinition) throw new IllegalUseOfConstraintAnnotationException();
	}
	
	private void weave(File classFolderOrClassFile) throws IOException {
		if (classFolderOrClassFile.isFile()) {
			try {
				classFileWeaver.weave(classFolderOrClassFile);
			} catch (NoWeavingNeededException e) {
				// do nothing.
			} catch (NotClassFileException e) {
				listener.skipped(classFolderOrClassFile);
			} catch (IllegalParameterOverridingException e) {
				listener.notified(classFolderOrClassFile, e);
				hasIllegalConstraintAnnotationDefinition = true;
			} catch (IllegalReturnValueOverridingException e) {
				listener.notified(classFolderOrClassFile, e);
				hasIllegalConstraintAnnotationDefinition = true;
			} catch (NoSuitableAnnotatedParameterTypeException e) {
				listener.notified(classFolderOrClassFile, e);
				hasIllegalConstraintAnnotationDefinition = true;
			} catch (NoSuitableAnnotatedReturnTypeException e) {
				listener.notified(classFolderOrClassFile, e);
				hasIllegalConstraintAnnotationDefinition = true;
			} catch (IllegalUseOfConstraintAnnotationException e) {
				throw new IllegalStateException("Unexpected IllegalUseOfConstraintAnnotationException detected.", e);
			}
			return;
		}
		File[] files = classFolderOrClassFile.listFiles();
		if (files == null) return;
		for (File child : files) {
			weave(child);
		}
	}
	
	// TODO: Change the design from observer to visitor.
	/**
	 * A listener interface for notifying waving status.
	 */
	public interface WeavingStatusListener {
		void notified(File classFile, IllegalParameterOverridingException e);
		void notified(File classFile, IllegalReturnValueOverridingException e);
		void notified(File classFile, NoSuitableAnnotatedParameterTypeException e);
		void notified(File classFile, NoSuitableAnnotatedReturnTypeException e);
		void skipped(File skippedFile);
	}
}
