/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.weaver;

import static jp.objectfanatics.assertion.weaver.WeaverUtil.assertProperlyOverridden;
import static jp.objectfanatics.assertion.weaver.WeaverUtil.getCtClass;
import static jp.objectfanatics.assertion.weaver.WeaverUtil.getTargetConstraintAssertionWeaver;
import static jp.objectfanatics.assertion.weaver.WeaverUtil.isAlreadyWoven;
import static jp.objectfanatics.assertion.weaver.WeaverUtil.isAnnotatedByConstraintAnnotation;
import static jp.objectfanatics.assertion.weaver.WeaverUtil.weaveWovenFlag;
import static jp.objectfanatics.commons.javassist.JavassistUtils.createClassPool;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getLineNumberEvenIfAvailableLineNumberInfoNotAvailable;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getParameterTypes;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getReturnType;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getUserDeclaredBehaviors;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getUserDeclaredMethods;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isAbstract;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isVoid;
import static jp.objectfanatics.commons.javassist.JavassistUtils.toAnnotationsAttribute;
import static jp.objectfanatics.commons.javassist.JavassistUtils.toBytecode;
import static jp.objectfanatics.commons.javassist.JavassistUtils.toParamValueIdentifier;
import static jp.objectfanatics.commons.javassist.JavassistUtils.toParameterAnnotationsAttribute;

import java.io.IOException;

import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.NotFoundException;
import javassist.bytecode.AnnotationsAttribute;
import javassist.bytecode.ParameterAnnotationsAttribute;
import javassist.bytecode.annotation.Annotation;
import jp.objectfanatics.assertion.weaver.constraint.MaxAssertionWeaver;
import jp.objectfanatics.assertion.weaver.constraint.MinAssertionWeaver;
import jp.objectfanatics.assertion.weaver.constraint.NotEmptyAssertionWeaver;
import jp.objectfanatics.assertion.weaver.constraint.NotNullAssertionWeaver;

// TODO: IllegalUseOfConstraintAnnotationException は、検査例外のほうがよいのでは？バグじゃないし。
// TODO: Weaverインタフェースを切り出して、こいつは WeaverImpl にしよう。見通しが悪すぎる。
// TODO: 全体的に、バイトコードから取得した変数名を利用するようにする。interface の場合は今までどおり。将来的には、annotation processor から収集した情報を利用しよう。
/**
 * Weaver for weaving assertion functions into class file bytes.
 * 
 * @author Makoto Sato
 */
public class Weaver {
	
	// TODO: こいつをコンストラクタのパラメータにすれば、ユーザーが自分用のアノテーションを好き勝手に作成できる。
	/**
	 * {@link ConstraintAssertionWeaver}s
	 */
	static final ConstraintAssertionWeaver[] constraintAssertionWeavers = {
		new NotNullAssertionWeaver(), new MinAssertionWeaver(), new MaxAssertionWeaver(), new NotEmptyAssertionWeaver()
	};
	
	// Checked!(9.10.2009)
	/**
	 * Extra classpaths for weaving.
	 */
	private final String[] extraClasspaths;
	
	// Checked!(9.10.2009)
	/**
	 * Initializes an instance.
	 * <p/>
	 * System class paths and <code>extraClasspaths</code> are used as classpaths for weaving.
	 * 
	 * @param extraClasspaths
	 *            extra classpaths.
	 */
	public Weaver(String... extraClasspaths) {
		this.extraClasspaths = extraClasspaths;
	}
	
	// Checked!(9.10.2009)
	/**
	 * Weaves assertion functions into class file bytes if needed.
	 * 
	 * @param classFileBytes
	 *            target class file bytes
	 * @return woven class file bytes
	 * @throws NoWeavingNeededException
	 *             if weaving is not needed.
	 * @throws IllegalStateException
	 *             if weaving is failed.
	 * @throws IllegalUseOfConstraintAnnotationException
	 *             if use of illegal constraint annotations are detected
	 */
	public byte[] weave(byte[] classFileBytes) throws NoWeavingNeededException, NotClassFileException {
		
		// notice : class pool must not be as an instance field because woven class, its frozen, can be referred again
		//          and throws frozen exception even just for getting information.
		ClassPool classPool = createClassPool(extraClasspaths);
		
		// get a handle for dealing with the target class file
		CtClass ctClass = getCtClass(classPool, classFileBytes);
		
		// weaves assertions
		return weave(ctClass);
	}
	
	/**
	 * Weaves assertion functions if needed.
	 * 
	 * @param ctClass
	 *            a handle for dealing with the target class file
	 * @return true if, and only if, assertion functions are woven into the class file bytes.
	 * @throws CannotCompileException
	 * @throws NotFoundException
	 * @throws IOException 
	 */
	private byte[] weave(CtClass ctClass) throws NoWeavingNeededException {
		// do nothing if assertion functions are already woven into class file bytes
		if (isAlreadyWoven(ctClass)) throw NoWeavingNeededException.SINGLETON;

		boolean isWoven = false;
		isWoven |= weaveAssertionsForParameters(ctClass);
		isWoven |= weaveAssertionsForReturnValue(ctClass);
		// weaves @Woven annotation into class file bytes if needed
		if (isWoven) {
			weaveWovenFlag(ctClass);
			return toBytecode(ctClass);
		}
		throw NoWeavingNeededException.SINGLETON;
	}


	
	/**
	 * Weaves assertion functions for parameters if needed.
	 * 
	 * @param ctClass
	 *            a handle for dealing with the target class file
	 * @return true if, and only if, assertion functions are woven into the class file bytes.
	 * @throws NotFoundException
	 * @throws CannotCompileException
	 */
	private boolean weaveAssertionsForParameters(CtClass ctClass) {
		boolean isWoven = false;
		for (CtBehavior ctBehavior : getUserDeclaredBehaviors(ctClass)) {
			isWoven |= weaveAssertionsForParameters(ctBehavior);
		}
		return isWoven;
	}
	
	/**
	 * Weaves assertion functions for parameters if needed.
	 * 
	 * @param ctBehavior
	 *            a handle for dealing with the target constructor or method
	 * @return true if, and only if, assertion functions are woven into the class file bytes.
	 * @throws NotFoundException
	 * @throws CannotCompileException
	 */
	private boolean weaveAssertionsForParameters(CtBehavior ctBehavior) {
		
			// return if ctBehavior is abstract
			if (isAbstract(ctBehavior)) return false;
			
			// return if there are no parameter annotations of constructors and methods
			ParameterAnnotationsAttribute attr = toParameterAnnotationsAttribute(ctBehavior);
			if (attr == null) return false;
			
			// weaves if needed
			boolean isWoven = false;
			for (int paramIndex = 0; paramIndex < getParameterTypes(ctBehavior).length; paramIndex++) {
				isWoven |= weaveAssertionsForAParameter(ctBehavior, paramIndex, attr.getAnnotations()[paramIndex]);
			}
			return isWoven;

	}
	
	/**
	 * Weaves assertion functions for a parameter if needed.
	 * 
	 * @param nonAbstractCtBehavior
	 *            a handle for dealing with the target constructor or method. It must be non-abstract.
	 * @param paramIndex
	 *            target parameter index
	 * @return true if, and only if, assertion functions are woven into the class file bytes.
	 * @throws NotFoundException
	 * @throws CannotCompileException
	 */
	private boolean weaveAssertionsForAParameter(CtBehavior nonAbstractCtBehavior, int paramIndex, Annotation[] parameterAnnotations) {
		// Use temp for clarity
		CtClass paramType = getParameterTypes(nonAbstractCtBehavior)[paramIndex];
		
		// weave assertions for an annotation in a parameter
		boolean isWoven = false;
		for (Annotation parameterAnnotation : parameterAnnotations) {
			isWoven |= weaveAssertionsForAnAnnotationInAParameter(nonAbstractCtBehavior, paramIndex, paramType, parameterAnnotation);
		}
		return isWoven;
	}
	
	/**
	 * Weaves assertion functions for an annotation in a parameter if needed.
	 * 
	 * @param nonAbstractCtBehavior
	 *            a handle for dealing with the target constructor or method. It must be non-abstract.
	 * @param paramIndex
	 *            target parameter index
	 * @param paramType
	 *            target parameter type
	 * @param parameterAnnotation
	 *            target parameter annotation
	 * @return true if, and only if, assertion functions are woven into the class file bytes.
	 * @throws CannotCompileException
	 */
	private boolean weaveAssertionsForAnAnnotationInAParameter(CtBehavior nonAbstractCtBehavior, int paramIndex, CtClass paramType, Annotation parameterAnnotation) {
		
		// ignore if annotation is not annotated by ConstraintAnnotation
		ClassPool classPool = nonAbstractCtBehavior.getDeclaringClass().getClassPool();
		if (!isAnnotatedByConstraintAnnotation(classPool, parameterAnnotation)) {
			return false;
		}
		else {
			ConstraintAssertionWeaver constraintAssertionWeaver = getTargetConstraintAssertionWeaver(parameterAnnotation);
			String paramValueIdentifier = toParamValueIdentifier(paramIndex);
			constraintAssertionWeaver.weaveParameterConstraintAssertion(nonAbstractCtBehavior, paramType, paramIndex, paramValueIdentifier, parameterAnnotation);
			return true;
		}
	}
	
	/**
	 * Weaves assertion functions for return values if needed.
	 * 
	 * @param ctClass
	 *            a handle for dealing with the target class file
	 * @return true if, and only if, assertion functions are woven into the class file bytes.
	 * @throws NotFoundException
	 * @throws CannotCompileException
	 */
	private boolean weaveAssertionsForReturnValue(CtClass ctClass) {
		boolean isWoven = false;
		for (CtMethod ctMethod : getUserDeclaredMethods(ctClass)) {
			assertProperlyOverridden(ctMethod);
			isWoven |= weaveAssertionsForReturnValue(ctMethod);
		}
		return isWoven;
	}
	
	/**
	 * Weaves assertion functions for return value if needed.
	 * 
	 * @param ctMethod
	 *            a handle for dealing with the target method
	 * @return true if, and only if, assertion functions are woven into the class file bytes.
	 * @throws NotFoundException
	 * @throws CannotCompileException
	 */
	private boolean weaveAssertionsForReturnValue(CtMethod ctMethod) {
		
		// returns if ctMethod is abstract
		if (isAbstract(ctMethod)) return false;
		
		// returns if there are no method annotations
		AnnotationsAttribute attr = toAnnotationsAttribute(ctMethod);
		if (attr == null) return false;
		
		// Use temp for clarity
		CtClass returnType = getReturnType(ctMethod);
		// TODO: void のチェックと例外送出は、ここでやったほうがよい。
		
		// weaves if needed
		boolean isWoven = false;
		for (Annotation annotation : attr.getAnnotations()) {
			isWoven |= weaveAssertionsForAnAnnotationInAReturnValue(ctMethod, annotation, returnType);
		}
		return isWoven;
	}
	
	private boolean weaveAssertionsForAnAnnotationInAReturnValue(CtMethod nonAbstractCtMethod, Annotation annotation, CtClass returnType) {
		// ignore if annotation is not annotated by ConstraintAnnotation
		ClassPool classPool = nonAbstractCtMethod.getDeclaringClass().getClassPool();
		if (!isAnnotatedByConstraintAnnotation(classPool, annotation)) {
			return false;
		}
		else {
			ConstraintAssertionWeaver constraintAssertionWeaver = getTargetConstraintAssertionWeaver(annotation);
			// TODO: void のチェックと例外送出は、上位メソッドでやったほうがよい。
			if (isVoid(returnType)) {
				int lineNumber = getLineNumberEvenIfAvailableLineNumberInfoNotAvailable(nonAbstractCtMethod);
				throw new NoSuitableAnnotatedReturnTypeException(constraintAssertionWeaver.getTargetConstraintAnnotation(), nonAbstractCtMethod, returnType, lineNumber);
			}
			else {
				constraintAssertionWeaver.weaveReturnTypeConstraintAssertion(nonAbstractCtMethod, returnType, annotation);
				return true;
			}
		}
	}
	
	/**
	 * Signals that weaving has failed.
	 */
	public static class NoWeavingNeededException extends Exception {
		public static final NoWeavingNeededException SINGLETON = new NoWeavingNeededException();
		private NoWeavingNeededException() {}
	}
	/**
	 * Signals that the bytes are not class file bytes.
	 */
	public static class NotClassFileException extends Exception {
		public NotClassFileException(Throwable e) { super(e); }
	}
	
	public static class IllegalParameterOverridingException extends IllegalUseOfConstraintAnnotationException {
		public final CtMethod overridingMethod;
		public final CtMethod overriddenMethod;
		public final Annotation annotation;
		public final int parameterIndex;
		public final String overridingParameterName;
		public final String overriddenParameterName;
		public final int overridingMethodLineNumber; // -1 if not available
		public final int overriddenMethodLineNumber; // -1 if not available
		public IllegalParameterOverridingException(CtMethod overridingMethod, CtMethod overriddenMethod, Annotation annotation, int parameterIndex, String overridingParameterName, String overriddenParameterName, int overridingMethodLineNumber, int overriddenMethodLineNumber) {
			super("Annotations of its ancestor's have to be annotated at " + overridingMethod.getLongName() + ". annotation = " + annotation.getTypeName() + ", ancestor = " + overriddenMethod.getLongName() + ", parameter index = " + parameterIndex + "(zero base).");
			this.overridingMethod = overridingMethod;
			this.overriddenMethod = overriddenMethod;
			this.annotation = annotation;
			this.parameterIndex = parameterIndex;
			this.overridingParameterName = overridingParameterName;
			this.overriddenParameterName = overriddenParameterName;
			this.overridingMethodLineNumber = overridingMethodLineNumber;
			this.overriddenMethodLineNumber = overriddenMethodLineNumber;
		}
	}

	public static class IllegalReturnValueOverridingException extends IllegalUseOfConstraintAnnotationException {
		public final CtMethod overridingMethod;
		public final CtMethod overriddenMethod;
		public final Annotation annotation;
		public final int overridingMethodLineNumber; // -1 if not available
		public final int overriddenMethodLineNumber; // -1 if not available
		public IllegalReturnValueOverridingException(CtMethod overridingMethod, CtMethod overriddenMethod, Annotation annotation, int overridingMethodLineNumber, int overriddenMethodLineNumber) {
			super("Annotations of its ancestor's have to be annotated at retrun value of " + overridingMethod.getLongName() + ". annotation = " + annotation.getTypeName() + ", ancestor = " + overriddenMethod.getLongName() + ".");
			this.overridingMethod = overridingMethod;
			this.overriddenMethod = overriddenMethod;
			this.annotation = annotation;
			this.overridingMethodLineNumber = overridingMethodLineNumber;
			this.overriddenMethodLineNumber = overriddenMethodLineNumber;
		}
	}
	public static class NoSuitableAnnotatedParameterTypeException extends IllegalUseOfConstraintAnnotationException {
		public final Class<? extends java.lang.annotation.Annotation> annotation;
		public final CtBehavior target;
		public final CtClass parameterType;
		public final int parameterIndex;
		public final String parameterName;
		public final int lineNumber; // -1 if not available
		public NoSuitableAnnotatedParameterTypeException(Class<? extends java.lang.annotation.Annotation> annotation, CtBehavior ctBehavior, CtClass parameterType, int parameterIndex, String parameterName, int lineNumber) {
			super(annotation.getName() + " annotation must not be used for a parameter which type is " + parameterType.getName() + ", at parameter index " + parameterIndex + "(zero base) in " + ctBehavior.getLongName());
			this.annotation = annotation;
			this.target = ctBehavior;
			this.parameterType = parameterType;
			this.parameterIndex = parameterIndex;
			this.parameterName = parameterName;
			this.lineNumber = lineNumber;
		}
	}	
	// TODO: create lineNumber here.
	public static class NoSuitableAnnotatedReturnTypeException extends IllegalUseOfConstraintAnnotationException {
		public final Class<? extends java.lang.annotation.Annotation> annotation;
		public final CtBehavior target;
		public final CtClass returnType;
		public final int lineNumber; // -1 if not available
		public NoSuitableAnnotatedReturnTypeException(Class<? extends java.lang.annotation.Annotation> annotation, CtMethod ctMethod, CtClass returnType, int lineNumber) {
			super(annotation.getName() + " annotation must not be used for return value which type is " + returnType.getName() + ", at " + ctMethod.getLongName());
			this.annotation = annotation;
			this.target = ctMethod;
			this.returnType = returnType;
			this.lineNumber = lineNumber;
		}
	}
	/**
	 * This exception is thrown when any use of illegal constraint annotations are detected.
	 */
	public static class IllegalUseOfConstraintAnnotationException extends RuntimeException {
		public IllegalUseOfConstraintAnnotationException(String message) {
			super(message);
		}
		
		public IllegalUseOfConstraintAnnotationException() {}
	}
}
