/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.weaver;

import static jp.objectfanatics.assertion.weaver.Weaver.constraintAssertionWeavers;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getAnnotations;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getAssignableTypesExpectItself;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getLineNumberEvenIfAvailableLineNumberInfoNotAvailable;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getParameterAnnotations;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getParameterNames;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getParameterTypes;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isAnnotatedWithMetaAnnotation;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isPrivate;
import static jp.objectfanatics.commons.javassist.JavassistUtils.toAnnotationsAttribute;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javassist.ClassPool;
import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.NotFoundException;
import javassist.bytecode.AnnotationsAttribute;
import javassist.bytecode.ClassFile;
import javassist.bytecode.ConstPool;
import javassist.bytecode.annotation.Annotation;
import jp.objectfanatics.assertion.constraints.ConstraintAnnotation;
import jp.objectfanatics.assertion.constraints.Woven;
import jp.objectfanatics.assertion.weaver.Weaver.IllegalParameterOverridingException;
import jp.objectfanatics.assertion.weaver.Weaver.IllegalReturnValueOverridingException;
import jp.objectfanatics.assertion.weaver.Weaver.NotClassFileException;
import jp.objectfanatics.commons.javassist.JavassistUtils.NoAvailableParameterNameInfoException;

public class WeaverUtil {
	
	/**
	 * Returns true if, and only if, assertion functions are already woven into the class file bytes.
	 * 
	 * @param ctClass
	 *            a handle for dealing with the target class file
	 * @return true if, and only if, assertion functions are already woven into the class file bytes
	 */
	static boolean isAlreadyWoven(CtClass ctClass) {
		AnnotationsAttribute attr = toAnnotationsAttribute(ctClass);
		if (attr == null) return false;
		return attr.getAnnotation(Woven.class.getName()) != null;
	}
	
	/**
	 * Weaves {@link Woven} annotation into given class file bytes.
	 * 
	 * @param ctClass
	 *            a handle for dealing with the target class file
	 * @return woven class file bytes
	 * @throws IOException
	 */
	static void weaveWovenFlag(CtClass ctClass) {
		ClassFile cf = ctClass.getClassFile();
		ConstPool cp = cf.getConstPool();
		AnnotationsAttribute attr = (AnnotationsAttribute) cf.getAttribute(AnnotationsAttribute.visibleTag);
		if (attr == null) {
			attr = new AnnotationsAttribute(cp, AnnotationsAttribute.visibleTag);
		}
		Annotation a = new Annotation(Woven.class.getName(), cp);
		attr.addAnnotation(a);
		cf.addAttribute(attr);
		cf.setVersionToJava5();
	}
	
	/**
	 * Returns true if, and only if, annotation is annotated by {@link ConstraintAnnotation}.
	 * 
	 * @param annotation
	 *            target
	 * @return true if, and only if, annotation is annotated by {@link ConstraintAnnotation}
	 */
	static boolean isAnnotatedByConstraintAnnotation(ClassPool classPool, Annotation annotation) {
		return isAnnotatedWithMetaAnnotation(classPool, annotation, ConstraintAnnotation.class.getName());
	}
	
	static void assertProperlyOverridden(CtMethod targetMethod) {
		ClassPool classPool = targetMethod.getDeclaringClass().getClassPool();
		for (CtMethod ancestorMethod : getAncestorMethods(targetMethod)) {
			assertParametersProperlyOverridden(classPool, targetMethod, ancestorMethod);
			assertReturnValueProperlyOverridden(classPool, targetMethod, ancestorMethod);
		}
	}
	
	/**
	 * Returns target {@link ConstraintAssertionWeaver} for the specified {@link Annotation}.
	 * 
	 * @param annotation
	 * @return target {@link ConstraintAssertionWeaver} for the specified {@link Annotation}.
	 */
	static ConstraintAssertionWeaver getTargetConstraintAssertionWeaver(Annotation annotation) {
		for (ConstraintAssertionWeaver constraintAssertionWeaver : constraintAssertionWeavers) {
			if (annotation.getTypeName().equals(constraintAssertionWeaver.getTargetConstraintAnnotation().getName())) {
				return constraintAssertionWeaver;
			}
		}
		throw new IllegalStateException(annotation.getTypeName() + " is annotated by " + ConstraintAnnotation.class.getName() + ", but no handler found.");
	}
	
	static Set<CtMethod> getAncestorMethods(CtMethod ctBehavior) {
		if (isPrivate(ctBehavior)) return new HashSet<CtMethod>();
		Set<CtMethod> ancestorMethods = new HashSet<CtMethod>();
		for (CtClass ancestorClassesOrInterface : getAssignableTypesExpectItself(ctBehavior)) {
			for (CtMethod candidate : ancestorClassesOrInterface.getMethods()) {
				if (isPrivate(candidate)) continue;
				if (!ctBehavior.getName().equals(candidate.getName())) continue;
				if (hasSameMethodParameterTypes(ctBehavior, candidate)) ancestorMethods.add(candidate);
			}
		}
		return ancestorMethods;
	}
	
	static boolean hasSameMethodParameterTypes(CtMethod targetMethod, CtMethod ancestorCandidateMethod) {
		CtClass[] targetMethodParameterTypes = getParameterTypes(targetMethod);
		CtClass[] ancestorCandidateMethodParameterTypes = getParameterTypes(ancestorCandidateMethod);
		if (targetMethodParameterTypes.length != ancestorCandidateMethodParameterTypes.length) return false;
		for (int i = 0; i < targetMethodParameterTypes.length; i++) {
			String targetMethodParameterTypeName = targetMethodParameterTypes[i].getName();
			String ancestorCandidateMethodParameterTypeName = ancestorCandidateMethodParameterTypes[i].getName();
			if (!targetMethodParameterTypeName.equals(ancestorCandidateMethodParameterTypeName)) {
				return false;
			}
		}
		return true;
	}
	
	static Annotation[] getParameterConstraintAnnotations(ClassPool classPool, CtBehavior ctBehavior, int paramIndex) {
		List<Annotation> parameterConstraintAnnotations = new ArrayList<Annotation>();
		for (Annotation parameterAnnotation : getParameterAnnotations(ctBehavior, paramIndex)) {
			if (isAnnotatedByConstraintAnnotation(classPool, parameterAnnotation)) parameterConstraintAnnotations.add(parameterAnnotation);
		}
		return parameterConstraintAnnotations.toArray(new Annotation[parameterConstraintAnnotations.size()]);
	}
	
	static Annotation[] getReturnValueConstraintAnnotations(ClassPool classPool, CtBehavior ctBehavior) {
		List<Annotation> retrunValueConstraintAnnotations = new ArrayList<Annotation>();
		for (Annotation annotation : getAnnotations(ctBehavior)) {
			if (isAnnotatedByConstraintAnnotation(classPool, annotation)) retrunValueConstraintAnnotations.add(annotation);
		}
		return retrunValueConstraintAnnotations.toArray(new Annotation[retrunValueConstraintAnnotations.size()]);
	}
	
	static void assertReturnValueProperlyOverridden(ClassPool classPool, CtMethod overridingMethod, CtMethod overriddenMethod) {
		for (Annotation ancestorAnnotation : getReturnValueConstraintAnnotations(classPool, overriddenMethod)) { // memo: if getParameterAnnotations(ancestorMethod, paramIndex) is heavy, use local temporary variable.
			Annotation[] annotations = getAnnotations(overridingMethod); // memo : if getParameterAnnotations(targetMethod) is heavy, targetMethodParameterAnnotations may be passed as a grand parent method parameter.
			boolean isExist = false;
			for (int k = 0; k < annotations.length; k++) {
				Annotation targetAnnotation = annotations[k];
				if (targetAnnotation.getTypeName().equals(ancestorAnnotation.getTypeName())) {
					isExist = true;// TODO: annotation parameter has to be checked. delegate to constraint assertion weaver.
					break;
				}
			}
			if (!isExist) {
				int overridingMethodLineNumber = getLineNumberEvenIfAvailableLineNumberInfoNotAvailable(overridingMethod);
				int overriddenMethodLineNumber = getLineNumberEvenIfAvailableLineNumberInfoNotAvailable(overriddenMethod);
				throw new IllegalReturnValueOverridingException(overridingMethod, overriddenMethod, ancestorAnnotation, overridingMethodLineNumber, overriddenMethodLineNumber);
			}
		}
	}
	
	static void assertParametersProperlyOverridden(ClassPool classPool, CtMethod targetMethod, CtMethod ancestorMethod) {
		for (int paramIndex = 0; paramIndex < getParameterTypes(targetMethod).length; paramIndex++) {
			assertParameterProperlyOverridden(classPool, targetMethod, ancestorMethod, paramIndex);
		}
	}
	static void assertParameterProperlyOverridden(ClassPool classPool, CtMethod targetMethod, CtMethod ancestorMethod, int paramIndex) {
		for (Annotation ancestorAnnotation : getParameterConstraintAnnotations(classPool, ancestorMethod, paramIndex)) { // memo: if getParameterAnnotations(ancestorMethod, paramIndex) is heavy, use local temporary variable.
			Annotation[] annotations = getParameterAnnotations(targetMethod)[paramIndex]; // memo : if getParameterAnnotations(targetMethod) is heavy, targetMethodParameterAnnotations may be passed as a grand parent method parameter.
			boolean isExist = false;
			for (int k = 0; k < annotations.length; k++) {
				Annotation targetAnnotation = annotations[k];
				if (targetAnnotation.getTypeName().equals(ancestorAnnotation.getTypeName())) {
					isExist = true;// TODO: annotation parameter has to be checked. delegate to constraint assertion weaver.
					break;
				}
			}
			if (!isExist) {
				String targetParamName = getParameterNamesEvenIfNoAvailableParameterNameInfo(targetMethod)[paramIndex];
				String ancestorParamName = getParameterNamesEvenIfNoAvailableParameterNameInfo(ancestorMethod)[paramIndex];
				int overridingMethodLineNumber = getLineNumberEvenIfAvailableLineNumberInfoNotAvailable(targetMethod);
				int overriddenMethodLineNumber = getLineNumberEvenIfAvailableLineNumberInfoNotAvailable(ancestorMethod);
				throw new IllegalParameterOverridingException(targetMethod, ancestorMethod, ancestorAnnotation, paramIndex, targetParamName, ancestorParamName, overridingMethodLineNumber, overriddenMethodLineNumber);
			}
		}
	}
	
	public static String[] getParameterNamesEvenIfNoAvailableParameterNameInfo(CtBehavior ctMethod) {
		try {
			return getParameterNames(ctMethod);
		}
		catch (NoAvailableParameterNameInfoException e) {}
		catch (NotFoundException e) {}
		String[] parameterNames = new String[getParameterTypes(ctMethod).length];
		for (int i = 0; i < parameterNames.length; i++) {
			parameterNames[i] = "" + i + "(zero base)";
		}
		return parameterNames;
	}

	/**
	 * Returns a handle for dealing with the target class file.
	 * 
	 * @param classfileBuffer
	 *            class file bytes
	 * @return a handle for dealing with the target class file
	 * @throws IOException
	 */
	public static CtClass getCtClass(ClassPool classPool, byte[] classfileBuffer) throws NotClassFileException {
		try {
			InputStream in = new ByteArrayInputStream(classfileBuffer);
			CtClass ctClass = classPool.makeClass(in);
			return ctClass;
		}
		catch (IOException e) {
			// TODO: 例外内容とハンドリング方法の再確認。
			throw new NotClassFileException(e);
		}
	}
}
