/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.weaver.constraint;

import static jp.objectfanatics.commons.javassist.JavassistUtils.isAbstract;
import javassist.CannotCompileException;
import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.bytecode.annotation.Annotation;
import jp.objectfanatics.assertion.weaver.ConstraintAssertionWeaver;

/**
 * Abstract class for {@link ConstraintAssertionWeaver}.
 * 
 * @author Makoto Sato
 */
abstract public class AbstractAssertionWeaver implements ConstraintAssertionWeaver {
	
	/**
	 * Target constraint annotation class.
	 */
	protected final Class<? extends java.lang.annotation.Annotation> targetConstraintAnnotation;
	
	/**
	 * Initializes a newly created <code>AbstractAssertionWeaver</code> object with <code>targetConstraintAnnotation</code>.
	 * 
	 * @param targetConstraintAnnotation
	 */
	public AbstractAssertionWeaver(Class<? extends java.lang.annotation.Annotation> targetConstraintAnnotation) {
		this.targetConstraintAnnotation = targetConstraintAnnotation;
	}
	
	@Override
	public Class<? extends java.lang.annotation.Annotation> getTargetConstraintAnnotation() {
		return targetConstraintAnnotation;
	}
	
	@Override
	public void weaveParameterConstraintAssertion(CtBehavior ctBehavior, CtClass paramType, int paramIndex, String paramValueIdentifier, Annotation annotation) {
		validateParameterConstraint(ctBehavior, paramType, paramIndex);
		if (isAbstract(ctBehavior)) return;
		doWeaveParameterConstraintAssertion(ctBehavior, paramType, paramIndex, paramValueIdentifier, annotation);
	}
	
	/**
	 * Validates parameter constraint annotation.
	 * 
	 * @param ctBehavior
	 * @param paramType
	 * @param paramIndex
	 */
	abstract protected void validateParameterConstraint(CtBehavior ctBehavior, CtClass paramType, int paramIndex);
	
	/**
	 * Weaves parameter constraint assertion.
	 * 
	 * @param ctBehavior
	 * @param paramType
	 * @param paramIndex
	 * @param paramValueIdentifier
	 * @param annotation
	 * @throws CannotCompileException
	 */
	abstract protected void doWeaveParameterConstraintAssertion(CtBehavior ctBehavior, CtClass paramType, int paramIndex, String paramValueIdentifier, Annotation annotation);
	
	@Override
	public void weaveReturnTypeConstraintAssertion(CtMethod ctMethod, CtClass returnType, Annotation annotation) {
		validateReturnTypeConstraint(ctMethod, returnType);
		if (isAbstract(ctMethod)) return;
		doWeaveReturnTypeConstraintAssertion(ctMethod, returnType, annotation);
	}
	
	/**
	 * Weaves return value constraint assertion.
	 * 
	 * @param ctMethod
	 * @param paramType
	 * @param annotation
	 * @throws CannotCompileException
	 */
	abstract protected void doWeaveReturnTypeConstraintAssertion(CtMethod ctMethod, CtClass paramType, Annotation annotation);
	
	/**
	 * Validates return value constraint annotation.
	 * 
	 * @param ctMethod
	 * @param returnType
	 */
	abstract protected void validateReturnTypeConstraint(CtMethod ctMethod, CtClass returnType);
	
	/**
	 * Generates assertion statement string.
	 * 
	 * @param assertionErrorCondition
	 * @param message
	 * @return assertion statement string
	 */
	private final static String generateAssrtionStatement(String assertionErrorCondition, String message, String exceptionName) {
		// using a constructor of AssertionError because assert statement is not supported by javassist
		return "if (" + assertionErrorCondition + ") throw new " + exceptionName + "(\"" + message + "\");";
	}
	
	/**
	 * Generates assertion statement string.
	 * 
	 * @param assertionErrorCondition
	 * @param message
	 * @return assertion statement string
	 */
	protected final static String generateParameterAssrtionStatement(String assertionErrorCondition, String message) {
		return generateAssrtionStatement(assertionErrorCondition, message, IllegalArgumentException.class.getSimpleName());
	}
	
	/**
	 * Generates assertion statement string.
	 * 
	 * @param assertionErrorCondition
	 * @param message
	 * @return assertion statement string
	 */
	protected final static String generateReturnValueAssrtionStatement(String assertionErrorCondition, String message) {
		return generateAssrtionStatement(assertionErrorCondition, message, IllegalStateException.class.getSimpleName());
	}
	
}
