/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.weaver.constraint;

import static jp.objectfanatics.commons.javassist.JavassistUtils.getLineNumberEvenIfAvailableLineNumberInfoNotAvailable;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isAtomicInteger;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isAtomicLong;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isBigDecimal;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isBigInteger;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isByte;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isDouble;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isFloat;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isInteger;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isLong;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isPrimitive;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isPrimitiveByte;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isPrimitiveDouble;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isPrimitiveFloat;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isPrimitiveInt;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isPrimitiveLong;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isPrimitiveShort;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isShort;
import javassist.CannotCompileException;
import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.bytecode.annotation.Annotation;
import javassist.bytecode.annotation.DoubleMemberValue;
import jp.objectfanatics.assertion.weaver.Weaver;
import jp.objectfanatics.assertion.weaver.WeaverUtil;
import jp.objectfanatics.assertion.weaver.Weaver.IllegalUseOfConstraintAnnotationException;

/**
 * Abstract class for number comparing weavers.
 * 
 * @author Makoto Sato
 */
abstract public class AbstractNumberComparingAssertionWeaver extends AbstractAssertionWeaver {
	
	/**
	 * for asserting Integer and Long value is not null.
	 */
	private NotNullAssertionWeaver notNullAssertionWeaver = new NotNullAssertionWeaver();
	
	/**
	 * Comparator.
	 */
	private final String comparator;
	
	/**
	 * Comparative word for number out of bounds.
	 */
	private final String comparativeWordForNumberOutOfbounds;
	
	/**
	 * Constructor.
	 * 
	 * @param targetConstraintAnnotation
	 */
	public AbstractNumberComparingAssertionWeaver(Class<? extends java.lang.annotation.Annotation> targetConstraintAnnotation, String comparator, String comparativeWordForNumberOutOfbounds) {
		// TODO: assertion;
		super(targetConstraintAnnotation);
		this.comparator = comparator;
		this.comparativeWordForNumberOutOfbounds = comparativeWordForNumberOutOfbounds;
	}
	
	/**
	 * Returns annotation presentation string.
	 * 
	 * @return annotation presentation string.
	 */
	protected String getAnnotationPresentationString() {
		return targetConstraintAnnotation.getSimpleName();
	}
	
	@Override
	protected void validateParameterConstraint(CtBehavior ctBehavior, CtClass paramType, int paramIndex) {
		// weaving is valid only for byte, short, int, long, float, double, Byte, Short, Integer, Long, Float, Double, BigInteger, BigDecimal, AtomicInteger, and AtomicLong.
		String parameterName = WeaverUtil.getParameterNamesEvenIfNoAvailableParameterNameInfo(ctBehavior)[paramIndex];
		int lineNumber = getLineNumberEvenIfAvailableLineNumberInfoNotAvailable(ctBehavior);
		if (!isValidType(paramType)) throw new Weaver.NoSuitableAnnotatedParameterTypeException(getTargetConstraintAnnotation(), ctBehavior, paramType, paramIndex, parameterName, lineNumber);
	}
	
	@Override
	protected void doWeaveParameterConstraintAssertion(CtBehavior ctBehavior, CtClass paramType, int paramIndex, String paramValueIdentifier, Annotation annotation) {
		try {
			// weaves
			String valueName = "parameter " + paramIndex + "(zero base)";
			String assertionStatement = createAssertionStatement(ctBehavior, paramType, paramValueIdentifier, annotation, valueName, true);
			ctBehavior.insertBefore(assertionStatement);
			
			// delegates @NotNull assertion
			if (!isPrimitive(paramType)) notNullAssertionWeaver.weaveParameterConstraintAssertion(ctBehavior, paramType, paramIndex, paramValueIdentifier, annotation);
		}
		catch (CannotCompileException e) {
			throw new IllegalStateException(e.getMessage(), e);
		}
	}
	
	@Override
	protected void validateReturnTypeConstraint(CtMethod ctMethod, CtClass returnType) {
		// weaving is valid only for int, long, Integer and Long
		int lineNumber = getLineNumberEvenIfAvailableLineNumberInfoNotAvailable(ctMethod);
		if (!isValidType(returnType)) throw new Weaver.NoSuitableAnnotatedReturnTypeException(getTargetConstraintAnnotation(), ctMethod, returnType, lineNumber);
	}
	
	@Override
	protected void doWeaveReturnTypeConstraintAssertion(CtMethod ctMethod, CtClass returnType, Annotation annotation) {
		try {
			// delegates @NotNull assertion
			if (!isPrimitive(returnType)) notNullAssertionWeaver.weaveReturnTypeConstraintAssertion(ctMethod, returnType, annotation);
			
			// weaves
			String valueName = "return value";
			String assertionStatement = createAssertionStatement(ctMethod, returnType, "$_", annotation, valueName, false);
			ctMethod.insertAfter(assertionStatement);
		}
		catch (CannotCompileException e) {
			throw new IllegalStateException(e.getMessage(), e);
		}
	}
	
	/**
	 * Asserts that <code>threshold</code> is in the range of byte.
	 * 
	 * @param ctBehavior
	 * @param valueName
	 * @param threshold
	 * @throws IllegalUseOfConstraintAnnotationException
	 *             when min is out of the range of byte.
	 */
	protected void assertByte(CtBehavior ctBehavior, String valueName, double threshold) {
		if (threshold > Byte.MAX_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, true, "Byte.MAX_VALUE[=" + Byte.MAX_VALUE + "]");
		if (threshold < Byte.MIN_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, false, "Byte.MIN_VALUE[=" + Byte.MIN_VALUE + "]");
	}
	
	/**
	 * Asserts that <code>threshold</code> is in the range of short.
	 * 
	 * @param ctBehavior
	 * @param valueName
	 * @param min
	 * @throws IllegalUseOfConstraintAnnotationException
	 *             when min is out of the range of short.
	 */
	protected void assertShort(CtBehavior ctBehavior, String valueName, double threshold) {
		if (threshold > Short.MAX_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, true, "Short.MAX_VALUE[=" + Short.MAX_VALUE + "]");
		if (threshold < Short.MIN_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, false, "Short.MIN_VALUE[=" + Short.MIN_VALUE + "]");
	}
	
	/**
	 * Asserts that <code>threshold</code> is in the range of int.
	 * 
	 * @param ctBehavior
	 * @param valueName
	 * @param threshold
	 * @throws IllegalUseOfConstraintAnnotationException
	 *             when threshold is out of the range of int.
	 */
	protected void assertInt(CtBehavior ctBehavior, String valueName, double threshold) {
		if (threshold > Integer.MAX_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, true, "Integer.MAX_VALUE[=" + Integer.MAX_VALUE + "]");
		if (threshold < Integer.MIN_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, false, "Integer.MIN_VALUE[=" + Integer.MIN_VALUE + "]");
	}
	
	/**
	 * Asserts that <code>threshold</code> is in the range of long.
	 * 
	 * @param ctBehavior
	 * @param valueName
	 * @param threshold
	 * @throws IllegalUseOfConstraintAnnotationException
	 *             when threshold is out of the range of long.
	 */
	protected void assertLong(CtBehavior ctBehavior, String valueName, double threshold) {
		if (threshold > Long.MAX_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, true, "Long.MAX_VALUE[=" + Long.MAX_VALUE + "]");
		if (threshold < Long.MIN_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, false, "Long.MIN_VALUE[=" + Long.MIN_VALUE + "]");
	}
	
	/**
	 * Asserts that <code>threshold</code> is in the range of float.
	 * 
	 * @param ctBehavior
	 * @param valueName
	 * @param threshold
	 * @throws IllegalUseOfConstraintAnnotationException
	 *             when threshold is out of the range of float.
	 */
	protected void assertFloat(CtBehavior ctBehavior, String valueName, double threshold) {
		if (threshold > Float.MAX_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, true, "Float.MAX_VALUE[=" + Float.MAX_VALUE + "]");
		if (threshold < -Float.MAX_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, false, "-Float.MAX_VALUE[=" + (-Float.MAX_VALUE) + "]");
	}
	
	/**
	 * Asserts that <code>threshold</code> is in the range of BigInteger.
	 * 
	 * @param ctBehavior
	 * @param valueName
	 * @param threshold
	 * @throws IllegalUseOfConstraintAnnotationException
	 *             when threshold is out of the range of BigInteger.
	 */
	protected void assertBigInteger(CtBehavior ctBehavior, String valueName, double threshold) {
		if (threshold > Long.MAX_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, true, "Long.MAX_VALUE[=" + Long.MAX_VALUE + "]");
		if (threshold < Long.MIN_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, false, "Long.MIN_VALUE[=" + Long.MIN_VALUE + "]");
	}
	
	/**
	 * Throws {@link IllegalUseOfConstraintAnnotationException} for threshold out of bounds.
	 * 
	 * @param ctBehavior
	 * @param valueName
	 * @param threshold
	 * @param isGreater
	 * @param target
	 */
	private void throwExceptionForThresholdOutOfBounds(CtBehavior ctBehavior, String valueName, double threshold, boolean isGreater, String target) {
		throw new IllegalUseOfConstraintAnnotationException("Specified value for " + valueName + " is " + ((isGreater ? "greater" : "less") + " than " + target) + " : @" + getAnnotationPresentationString() + "(" + threshold + ") : " + ctBehavior.getLongName());
	}
	
	private String createAssertionStatement(CtBehavior ctBehavior, CtClass type, String valueIdentifier, Annotation annotation, String valueName, boolean isParam) throws CannotCompileException {
		double threshold = DoubleMemberValue.class.cast(annotation.getMemberValue("value")).getValue();
		if (Double.isNaN(threshold) || Double.isInfinite(threshold)) throw new IllegalUseOfConstraintAnnotationException("" + threshold + " is not supported by @" + getAnnotationPresentationString() + " at " + ctBehavior.getLongName());
		String assertionErrorConditionJavaSource;
		// byte
		if (isPrimitiveByte(type)) {
			assertByte(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + " " + comparator + " " + threshold;
		}
		// Byte
		else if (isByte(type)) {
			assertByte(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + ".byteValue() " + comparator + " " + threshold;
		}
		// short
		else if (isPrimitiveShort(type)) {
			assertShort(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + " " + comparator + " " + threshold;
		}
		// Short
		else if (isShort(type)) {
			assertShort(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + ".shortValue() " + comparator + " " + threshold;
		}
		// int
		else if (isPrimitiveInt(type)) {
			assertInt(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + " " + comparator + " " + threshold;
		}
		// Integer, AtomicInteger
		else if (isInteger(type) || isAtomicInteger(type)) {
			assertInt(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + ".intValue() " + comparator + " " + threshold;
		}
		// long
		else if (isPrimitiveLong(type)) {
			assertLong(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + " " + comparator + " (double)" + threshold;
		}
		// Long, AtomicLong
		else if (isLong(type) || isAtomicLong(type)) {
			assertLong(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + ".longValue() " + comparator + " (double)" + threshold;
		}
		// float
		else if (isPrimitiveFloat(type)) {
			assertFloat(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + " " + comparator + " (double)" + threshold;
		}
		// Float
		else if (isFloat(type)) {
			assertFloat(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + ".floatValue() " + comparator + " (double)" + threshold;
		}
		// double
		else if (isPrimitiveDouble(type)) {
			assertionErrorConditionJavaSource = valueIdentifier + " " + comparator + " (double)" + threshold;
		}
		// Double
		else if (isDouble(type)) {
			assertionErrorConditionJavaSource = valueIdentifier + ".doubleValue() " + comparator + " (double)" + threshold;
		}
		// BigInteger
		else if (isBigInteger(type)) {
			assertBigInteger(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + ".longValue() " + comparator + " (double)" + threshold;
		}
		// BigDecimal
		else if (isBigDecimal(type)) {
			assertionErrorConditionJavaSource = valueIdentifier + ".doubleValue() " + comparator + " (double)" + threshold;
		}
		// Unexpected type
		else {
			throw new IllegalStateException("Unexpected type : " + type.getName());
		}
		String message = valueName + "[=\"+" + valueIdentifier + "+\"] is " + comparativeWordForNumberOutOfbounds + " than " + threshold + "."; // TODO: use parameter name in the message.
		String assrtionStatement = isParam ? generateParameterAssrtionStatement(assertionErrorConditionJavaSource, message) : generateReturnValueAssrtionStatement(assertionErrorConditionJavaSource, message);
		return assrtionStatement;
	}
	
	// TODO: isPrimitiveは boolean とか入るので付加。isPrimitiveNumberとか作成しよう。
	/**
	 * Returns true if, and only if, the specified type is byte, short, int, long, float, double, Byte, Short, Integer,
	 * Long, Float, Double, BigInteger, BigDecimal, AtomicInteger or AtomicLong.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is byte, short, int, long, float, double, Byte, Short, Integer,
	 *         Long, Float, Double, BigInteger, BigDecimal, AtomicInteger or AtomicLong.
	 */
	public static boolean isValidType(CtClass type) {
		return isPrimitive(type)         // primitive types
				|| isByte(type)          // java.lang.Byte
				|| isShort(type)         // java.lang.Short
				|| isInteger(type)       // java.lang.Integer
				|| isLong(type)          // java.lang.Long
				|| isFloat(type)         // java.lang.Float
				|| isDouble(type)        // java.lang.Double
				|| isBigInteger(type)    // java.math.BigInteger
				|| isBigDecimal(type)    // java.math.BigDecimal
				|| isAtomicInteger(type) // java.util.concurrent.atomic.AtomicInteger
				|| isAtomicLong(type);   // java.util.concurrent.atomic.AtomicLong
	}
}
