/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.weaver.constraint;

import static jp.objectfanatics.commons.javassist.JavassistUtils.getLineNumberEvenIfAvailableLineNumberInfoNotAvailable;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isAssignableToCharSequence;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isAssignableToCollection;
import javassist.CannotCompileException;
import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.bytecode.annotation.Annotation;
import jp.objectfanatics.assertion.constraints.NotEmpty;
import jp.objectfanatics.assertion.constraints.NotNull;
import jp.objectfanatics.assertion.weaver.ConstraintAssertionWeaver;
import jp.objectfanatics.assertion.weaver.Weaver;
import jp.objectfanatics.assertion.weaver.WeaverUtil;

/**
 * Implementation of {@link ConstraintAssertionWeaver} for @{@link NotEmpty} annotation.
 * 
 * @author Makoto Sato
 */
public class NotEmptyAssertionWeaver extends AbstractAssertionWeaver {
	
	/**
	 * for asserting the value is not null.
	 */
	private NotNullAssertionWeaver notNullAssertionWeaver = new NotNullAssertionWeaver();
	
	/**
	 * Constructor.
	 */
	public NotEmptyAssertionWeaver() {
		super(NotEmpty.class);
	}
	
	@Override
	protected void validateParameterConstraint(CtBehavior ctBehavior, CtClass paramType, int paramIndex) {
		// weaving is valid only for a type compatible with CharSequence, Collection, or array
		String parameterName = WeaverUtil.getParameterNamesEvenIfNoAvailableParameterNameInfo(ctBehavior)[paramIndex];
		int lineNumber = getLineNumberEvenIfAvailableLineNumberInfoNotAvailable(ctBehavior);
		if (!isValidType(paramType)) throw new Weaver.NoSuitableAnnotatedParameterTypeException(getTargetConstraintAnnotation(), ctBehavior, paramType, paramIndex, parameterName, lineNumber);
	}
	
	@Override
	protected void doWeaveParameterConstraintAssertion(CtBehavior ctBehavior, CtClass paramType, int paramIndex, String paramValueIdentifier, Annotation annotation) {
		try {
		// weaves
		String valueName = "parameter " + paramIndex + "(zero base)";
		String assertionStatement = createAssertionStatement(ctBehavior, paramType, paramValueIdentifier, annotation, valueName, true);
		ctBehavior.insertBefore(assertionStatement);
		
		// delegates @NotNull assertion
		notNullAssertionWeaver.weaveParameterConstraintAssertion(ctBehavior, paramType, paramIndex, paramValueIdentifier, annotation);
		} catch (CannotCompileException e) {
			throw new IllegalStateException(e.getMessage(), e);
		}
	}
	
	@Override
	protected void validateReturnTypeConstraint(CtMethod ctMethod, CtClass returnType) {
		// weaving is valid only for a type compatible with CharSequence, Collection, or array
		int lineNumber = getLineNumberEvenIfAvailableLineNumberInfoNotAvailable(ctMethod);
		if (!isValidType(returnType)) throw new Weaver.NoSuitableAnnotatedReturnTypeException(getTargetConstraintAnnotation(), ctMethod, returnType, lineNumber);
	}
	
	@Override
	protected void doWeaveReturnTypeConstraintAssertion(CtMethod ctMethod, CtClass returnType, Annotation annotation) {
		
		try {
			// delegates @NotNull assertion
			notNullAssertionWeaver.weaveReturnTypeConstraintAssertion(ctMethod, returnType, annotation);
			
			// weaves
			String valueName = "return value";
			String assertionStatement = createAssertionStatement(ctMethod, returnType, "$_", annotation, valueName, false);
			ctMethod.insertAfter(assertionStatement);
		}
		catch (CannotCompileException e) {
			throw new IllegalStateException(e.getMessage(), e);
		}
	}
	
	@NotEmpty
	private String createAssertionStatement(@NotNull CtBehavior ctBehavior, @NotNull CtClass type, @NotEmpty String valueIdentifier, @NotNull Annotation annotation, @NotEmpty String valueName, boolean isParam) {
		String assertionErrorCondition;
		if (isAssignableToCharSequence(type)) {
			assertionErrorCondition = valueIdentifier + ".length() <= 0";
		}
		else if (type.isArray()) {
			assertionErrorCondition = valueIdentifier + ".length <= 0";
		}
		else if (isAssignableToCollection(type)) {
			assertionErrorCondition = valueIdentifier + ".size() <= 0";
		}
		else {
			throw new IllegalStateException("Unexpected type : " + type.getName());
		}
		String message = valueName + " is empty."; // TODO: use parameter name in the message.
		String assrtionStatement = isParam ? generateParameterAssrtionStatement(assertionErrorCondition, message) : generateReturnValueAssrtionStatement(assertionErrorCondition, message);
		return assrtionStatement;
	}
	
	/**
	 * Returns true if, and only if, the parameter type is compatible with CharSequence, Collection, or array.
	 * 
	 * @param paramType
	 *            target parameter type
	 * @return true if, and only if, the parameter type is compatible with CharSequence, Collection, or array.
	 */
	private boolean isValidType(@NotNull CtClass paramType) {
		return isAssignableToCharSequence(paramType) || paramType.isArray() || isAssignableToCollection(paramType);
	}
}
