/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.commons.javassist;

import java.io.IOException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;

import javassist.CannotCompileException;
import javassist.ClassPath;
import javassist.ClassPool;
import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.Modifier;
import javassist.NotFoundException;
import javassist.bytecode.AnnotationsAttribute;
import javassist.bytecode.CodeAttribute;
import javassist.bytecode.LineNumberAttribute;
import javassist.bytecode.LocalVariableAttribute;
import javassist.bytecode.ParameterAnnotationsAttribute;
import javassist.bytecode.annotation.Annotation;

// TODO: groupId=jp.objectfanatics.commons, artifactId=javassist として単独のアーティファクトに切り出しましょう。
public class JavassistUtils {
	
	/**
	 * Returns {@link AnnotationsAttribute} annotating for the ctClass.
	 * 
	 * @param ctClass
	 *            {@link CtClass}
	 * @return {@link AnnotationsAttribute} annotating for the ctClass
	 */
	public static AnnotationsAttribute toAnnotationsAttribute(CtClass ctClass) {
		return (AnnotationsAttribute) ctClass.getClassFile().getAttribute(AnnotationsAttribute.visibleTag);
	}
	
	/**
	 * Returns {@link AnnotationsAttribute} annotating for the ctMethod.
	 * 
	 * @param ctBehavior
	 *            {@link CtMethod}
	 * @return {@link AnnotationsAttribute} annotating for the ctMethod
	 */
	public static AnnotationsAttribute toAnnotationsAttribute(CtBehavior ctBehavior) {
		return (AnnotationsAttribute) ctBehavior.getMethodInfo().getAttribute(AnnotationsAttribute.visibleTag);
	}
	
	/**
	 * Returns {@link ParameterAnnotationsAttribute} of the specified {@link CtBehavior}.
	 * 
	 * @param ctBehavior
	 *            {@link CtBehavior}
	 * @return {@link ParameterAnnotationsAttribute} of the specified {@link CtBehavior}
	 */
	public static ParameterAnnotationsAttribute toParameterAnnotationsAttribute(CtBehavior ctBehavior) {
		return (ParameterAnnotationsAttribute) ctBehavior.getMethodInfo().getAttribute(ParameterAnnotationsAttribute.visibleTag);
	}
	
	public static Annotation[][] getParameterAnnotations(CtBehavior ctBehavior) {
		ParameterAnnotationsAttribute parameterAnnotationAttribute = toParameterAnnotationsAttribute(ctBehavior);
		return (parameterAnnotationAttribute != null) ? parameterAnnotationAttribute.getAnnotations() : new Annotation[getParameterTypes(ctBehavior).length][0];
	}
	
	public static Annotation[] getParameterAnnotations(CtBehavior ctBehavior, int paramIndex) {
		return getParameterAnnotations(ctBehavior)[paramIndex];
	}
	
	public static Annotation[] getAnnotations(CtBehavior ctBehavior) {
		AnnotationsAttribute attr = toAnnotationsAttribute(ctBehavior);
		return attr != null ? attr.getAnnotations() : new Annotation[0];
	}
	
	// TODO: クラスやインタフェース互換性関連のユーティリティメソッドは、全体的に整合性を考えて、名前を十分に検討し、整理しましょう。
	public static CtClass[] getAncestorClasses(CtClass ctClass) {
		Set<CtClass> ancestorClassess = new LinkedHashSet<CtClass>();
		CtClass current = ctClass;
		while (true) {
			CtClass superClass = getSuperClass(current);
			if (superClass == null) return ancestorClassess.toArray(new CtClass[ancestorClassess.size()]);
			ancestorClassess.add(superClass);
			current = superClass;
		}
	}
	
	// TODO: クラスやインタフェース互換性関連のユーティリティメソッドは、全体的に整合性を考えて、名前を十分に検討し、整理しましょう。
	// TODO: refine and unit test.
	public static Set<CtClass> getAssignableTypesExpectItself(CtMethod ctMethod) {
		Set<CtClass> assignableSuperTypes = new LinkedHashSet<CtClass>();
		addAll(assignableSuperTypes, getAncestorClasses(ctMethod.getDeclaringClass()));
		addAll(assignableSuperTypes, getAssignableInterfacesExceptItself(ctMethod.getDeclaringClass()));
		return assignableSuperTypes;
	}
	
	private static <T> void addAll(Collection<T> collection, T[] array) {
		for (T element : array)
			collection.add(element);
	}
	
	/**
	 * Creates class pool with system classpaths and extra classpaths.
	 * 
	 * @param extraClasspaths
	 *            extra classpaths
	 * @return a class pool configured with system classpaths and extra classpaths
	 */
	public static ClassPool createClassPool(String[] extraClasspaths) {
		ClassPool classPool = new ClassPool();
		classPool.appendSystemPath();
		for (String classpath : extraClasspaths) {
			appendClassPath(classPool, classpath);
		}
		return classPool;
	}
	
	public static AnnotationsAttribute toAnnotationsAttribute(ClassPool classPool, Annotation annotation) throws NotFoundException {
		CtClass annotationCtClass = classPool.get(annotation.getTypeName());
		return JavassistUtils.toAnnotationsAttribute(annotationCtClass);
	}
	
	public static boolean isAnnotatedWithMetaAnnotation(ClassPool classPool, Annotation annotation, String metaAnnotationFqcn) {
		try {
			AnnotationsAttribute annotationAttribute = toAnnotationsAttribute(classPool, annotation);
			if (annotationAttribute == null) return false;
			Annotation metaAnnotation = annotationAttribute.getAnnotation(metaAnnotationFqcn);
			return metaAnnotation != null;
		} catch (NotFoundException e) {
			throw new IllegalArgumentException(annotation.getTypeName() + " is not found in classpath. Please check your classpath.");
		}
	}
	


	// TODO: Check
	/**
	 * Returns {@link CtBehavior}s that each of them can be returned with
	 * {@link CtClass#getDeclaredBehavior(String, CtClass[])}. {@link CtClass#getDeclaredBehaviors()} returns multiple
	 * methods when an overriding method has one or more different return types.
	 * 
	 * @param ctClass
	 * @return
	 * @throws NotFoundException
	 */
	public static CtBehavior[] getUserDeclaredBehaviors(CtClass ctClass) {
		List<CtBehavior> userDeclaredBehaviors = new ArrayList<CtBehavior>();
		// methods
		for (CtMethod ctMethod : getUserDeclaredMethods(ctClass)) {
			userDeclaredBehaviors.add(ctMethod);
		}
		// constructors
		for (CtBehavior declaredConstructor : ctClass.getDeclaredConstructors()) {
			userDeclaredBehaviors.add(declaredConstructor);
		}
		return userDeclaredBehaviors.toArray(new CtBehavior[userDeclaredBehaviors.size()]);
	}

	/**
	 * Returns {@link CtMethod}s that each of them can be returned with
	 * {@link CtClass#getDeclaredMethod(String, CtClass[])}. {@link CtClass#getDeclaredMethods()} returns multiple
	 * methods when an overriding method has one or more different return types.
	 * 
	 * @param ctClass
	 * @return
	 */
	public static CtMethod[] getUserDeclaredMethods(CtClass ctClass) {
			List<CtMethod> userDeclaredMethods = new ArrayList<CtMethod>();
			for (CtMethod declaredMethod : ctClass.getDeclaredMethods()) {
				String methodName = declaredMethod.getName();
				CtClass[] parameterTypes = getParameterTypes(declaredMethod);
				CtMethod userDeclaredMethod = getDeclaredMethod(ctClass, methodName, parameterTypes);
				String ctMethodReturnTypeName = getReturnType(declaredMethod).getName();
				String userDeclaredMethodReturnTypeName = getReturnType(userDeclaredMethod).getName();
				if (ctMethodReturnTypeName.equals(userDeclaredMethodReturnTypeName)) {
					userDeclaredMethods.add(declaredMethod);
				}
			}
			return userDeclaredMethods.toArray(new CtMethod[userDeclaredMethods.size()]);
	}





	// TODO: Check and move JavassitUtil.java
	public static int getLineNumber(CtBehavior ctBehavior) throws LineNumberNotAvailableException {
		CodeAttribute codeAttribute = ctBehavior.getMethodInfo().getCodeAttribute();
		if (codeAttribute == null) throw LineNumberNotAvailableException.SINGLETON;
		LineNumberAttribute attr = (LineNumberAttribute) codeAttribute.getAttribute(LineNumberAttribute.tag);
		if (attr == null) throw LineNumberNotAvailableException.SINGLETON;
		return attr.lineNumber(0);
	}
	
	// TODO: Check and move JavassitUtil.java
	public static class LineNumberNotAvailableException extends Exception {
		public static final LineNumberNotAvailableException SINGLETON = new LineNumberNotAvailableException();
		private LineNumberNotAvailableException() {}
	}

	public static String[] getParameterNames(CtBehavior ctMethod) throws NotFoundException, NoAvailableParameterNameInfoException {
		String[] parameterNames = new String[ctMethod.getParameterTypes().length];
		CodeAttribute codeAttribute = ctMethod.getMethodInfo().getCodeAttribute();
		if (codeAttribute == null) {
			throw NoAvailableParameterNameInfoException.CODE_ATTRIBUTE_NOT_FOUND;
		}
		LocalVariableAttribute locals = (LocalVariableAttribute) codeAttribute.getAttribute("LocalVariableTable");
		if (locals == null) {
			throw NoAvailableParameterNameInfoException.LOCAL_VARIABLE_ATTRIBUTE_NOT_FOUND;
		}
		for (int i = 0, index = 0; i < parameterNames.length; index++) {
			if ("this".equals(locals.variableName(index))) continue;
			parameterNames[i++] = locals.variableName(index);
		}
		return parameterNames;
	}
	
	public static class NoAvailableParameterNameInfoException extends Exception {
		public static final NoAvailableParameterNameInfoException CODE_ATTRIBUTE_NOT_FOUND = new NoAvailableParameterNameInfoException("Code attribute not found.");
		public static final NoAvailableParameterNameInfoException LOCAL_VARIABLE_ATTRIBUTE_NOT_FOUND = new NoAvailableParameterNameInfoException("LocalVariableAttribute not found.");
		private NoAvailableParameterNameInfoException(String message) {}
	}
	
	public static int getLineNumberEvenIfAvailableLineNumberInfoNotAvailable(CtBehavior ctBehavior) {
		try {
			return getLineNumber(ctBehavior);
		}
		catch (LineNumberNotAvailableException e) {
			return -1;
		}
	}
	
	
	
	
	
	/**
	 * Returns true if, and only if, the type of ctClass is void.
	 * 
	 * @param ctClass
	 * @return true if, and only if, return type is void
	 */
	public static boolean isVoid(CtClass ctClass) {
		return ctClass.isPrimitive() && "void".equals(ctClass.getName());
	}
	
	public static boolean isPrivate(CtMethod targetMethod) {
		return Modifier.isPrivate(targetMethod.getModifiers());
	}
	
	public static boolean isAbstract(CtBehavior ctBehavior) {
		return Modifier.isAbstract(ctBehavior.getModifiers());
	}
	
	public static String getSourceFileName(CtClass ctClass) {
		return ctClass.getClassFile().getSourceFile();
	}
	
	public static String getSourceFileName(CtBehavior behavior) {
		return getSourceFileName(behavior.getDeclaringClass());
	}
	
	// Checked!(9.11.2009)
	/**
	 * Returns a parameter value identifier for the specified parameter index. It is used in the source code to be compiled by javassist.
	 * 
	 * @param paramIndex target parameter index
	 * @return a parameter value identifier for the specified parameter index.
	 */
	public static String toParamValueIdentifier(int paramIndex) {
		return "$" + (paramIndex + 1);
	}
	
	
	
	
	// --------------------------------------------------------------------------------------------
	// Type testers
	// --------------------------------------------------------------------------------------------
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is primitive boolean.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is primitive boolean.
	 */
	public static boolean isPrimitiveBoolean(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return boolean.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is primitive char.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is primitive char.
	 */
	public static boolean isPrimitiveChar(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return char.class.getName().equals(type.getName());
	}

	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is primitive byte.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is primitive byte.
	 */
	public static boolean isPrimitiveByte(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return byte.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is short.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is short.
	 */
	public static boolean isPrimitiveShort(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return short.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is int.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is int.
	 */
	public static boolean isPrimitiveInt(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return int.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is long.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is long.
	 */
	public static boolean isPrimitiveLong(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return long.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is float.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is float.
	 */
	public static boolean isPrimitiveFloat(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return float.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is double.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is double.
	 */
	public static boolean isPrimitiveDouble(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return double.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is Boolean.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is Boolean.
	 */
	public static boolean isBoolean(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return Boolean.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is Character.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is Character.
	 */
	public static boolean isCharacter(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return Character.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is Byte.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is Byte.
	 */
	public static boolean isByte(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return Byte.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is Short.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is Short.
	 */
	public static boolean isShort(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return Short.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is Integer.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is Integer.
	 */
	public static boolean isInteger(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return Integer.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is AtomicInteger.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is AtomicInteger.
	 */
	public static boolean isAtomicInteger(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return AtomicInteger.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is Long.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is Long.
	 */
	public static boolean isLong(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return Long.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is AtomicLong.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is AtomicLong.
	 */
	public static boolean isAtomicLong(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return AtomicLong.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is Float.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is Float.
	 */
	public static boolean isFloat(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return Float.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is Double.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is Double.
	 */
	public static boolean isDouble(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return Double.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is BigInteger.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is BigInteger.
	 */
	public static boolean isBigInteger(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return BigInteger.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is BigDecimal.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is BigDecimal.
	 */
	public static boolean isBigDecimal(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return BigDecimal.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is compatible with CharSequence.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is compatible with CharSequence.
	 */
	public static boolean isAssignableToCharSequence(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return isSelfOrImplementor(type, CharSequence.class);
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is compatible with java.util.Collection.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is compatible with java.util.Collection.
	 */
	public static boolean isAssignableToCollection(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return isSelfOrImplementor(type, Collection.class);
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is primitive.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is primitive
	 */
	public static boolean isPrimitive(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return isPrimitiveBoolean(type) // boolean
			|| isPrimitiveChar(type)    // char
			|| isPrimitiveByte(type)    // byte
			|| isPrimitiveShort(type)   // short
			|| isPrimitiveInt(type)     // int
			|| isPrimitiveLong(type)    // long
			|| isPrimitiveFloat(type)   // float
			|| isPrimitiveDouble(type); // double
	}
	
	// TODO: クラスやインタフェース互換性関連のユーティリティメソッドは、全体的に整合性を考えて、名前を十分に検討し、整理しましょう。
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is compatible with the interface type expressed by <code>interfaceTypeFqcn</code>.
	 * 
	 * @param type
	 *            target type
	 * @param interfaceTypeFqcn
	 *            FQCN of interface type
	 * @return true if, and only if, the specified type is compatible with the interface type expressed by <code>interfaceTypeFqcn</code>
	 */
	public static boolean isSelfOrImplementor(CtClass type, String interfaceTypeFqcn) {
		// null check
		if (type == null) throw new IllegalArgumentException("type is null.");
		if (interfaceTypeFqcn == null) throw new IllegalArgumentException("interfaceTypeFqcn is null.");
		// self
		if (type.isInterface() && interfaceTypeFqcn.equals(type.getName())) return true;
		// for implementor
		for (CtClass implementedInterface : getAssignableInterfacesExceptItself(type)) {
			if (interfaceTypeFqcn.equals(implementedInterface.getName())) return true;
		}
		// no match
		return false;
	}
	
	// TODO: クラスやインタフェース互換性関連のユーティリティメソッドは、全体的に整合性を考えて、名前を十分に検討し、整理しましょう。
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns all assignable interfaces except for itself.
	 * 
	 * @param ctClass
	 *            target {@link CtClass}
	 * @return all assignable interfaces except for itself
	 * @throws IllegalArgumentException
	 *             if <code>ctClass</code> is null
	 * @throws IllegalStateException
	 *             if NotFoundException is thrown
	 */
	public static CtClass[] getAssignableInterfacesExceptItself(CtClass ctClass) {
		if (ctClass == null) throw new IllegalArgumentException("ctClass is null.");
		Set<CtClass> result = new HashSet<CtClass>();
		for (CtClass superInterface : getInterfaces(ctClass)) {
			result.add(superInterface);
			addAll(result, getAssignableInterfacesExceptItself(superInterface));
		}
		for (CtClass ancestorClass : getAncestorClasses(ctClass)) {
			for (CtClass superInterface : getInterfaces(ancestorClass)) {
				result.add(superInterface);
				addAll(result, getAssignableInterfacesExceptItself(superInterface));
			}
		}
		return result.toArray(new CtClass[result.size()]);
	}
	
	// TODO: クラスやインタフェース互換性関連のユーティリティメソッドは、全体的に整合性を考えて、名前を十分に検討し、整理しましょう。
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is assignable to the interface type expressed by <code>interfaceTypeFqcn</code>.
	 * 
	 * @param type
	 *            target type
	 * @param clazz
	 *            interface type
	 * @return true if, and only if, the specified type is assignable to the interface type expressed by <code>interfaceTypeFqcn</code>
	 */
	public static boolean isSelfOrImplementor(CtClass type, Class<?> interfaceType) {
		// null check
		if (type == null) throw new IllegalArgumentException("type is null.");
		if (interfaceType == null) throw new IllegalArgumentException("interfaceType is null.");
		// non-interface
		if (!interfaceType.isInterface()) throw new IllegalArgumentException("interfaceType is not interface : " + interfaceType.getName());
		// delegate
		return isSelfOrImplementor(type, interfaceType.getName());
	}
	
	// --------------------------------------------------------------------------------------------
	// Checked exception suppressors
	// --------------------------------------------------------------------------------------------
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Checked exception suppressor for {@link CtClass#toBytecode()}.
	 * </p>
	 * This method converts {@link NotFoundException} to {@link IllegalStateException}.
	 * 
	 * @param ctClass
	 *            target {@link CtClass}
	 * @return return value of {@link CtClass#toBytecode()}
	 * @throws IllegalArgumentException
	 *             if <code>ctClass</code> is null.
	 * @throws IllegalStateException
	 *             if NotFoundException is thrown
	 */
	public static byte[] toBytecode(CtClass ctClass) {
		if (ctClass == null) throw new IllegalArgumentException("ctClass is null.");
		try {
			return ctClass.toBytecode();
		}
		catch (IOException e) {
			throw new IllegalStateException(e.getMessage(), e);
		}
		catch (CannotCompileException e) {
			throw new IllegalStateException(e.getMessage(), e);
		}
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Checked exception suppressor for {@link CtBehavior#getParameterTypes()}.
	 * </p>
	 * This method converts {@link NotFoundException} to {@link IllegalStateException}.
	 * 
	 * @param ctBehavior
	 *            target {@link CtBehavior}
	 * @return return value of {@link CtBehavior#getParameterTypes()}
	 * @throws IllegalArgumentException
	 *             if <code>ctBehavior</code> is null
	 * @throws IllegalStateException
	 *             if NotFoundException is thrown
	 */
	public static CtClass[] getParameterTypes(CtBehavior ctBehavior) {
		if (ctBehavior == null) throw new IllegalArgumentException("ctBehavior is null.");
		try {
			return ctBehavior.getParameterTypes();
		}
		catch (NotFoundException e) {
			throw new IllegalStateException(e.getMessage(), e);
		}
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Checked exception suppressor for {@link CtMethod#getReturnType()}.
	 * </p>
	 * This method converts {@link NotFoundException} to {@link IllegalStateException}.
	 * 
	 * @param ctMethod
	 *            target {@link CtMethod}
	 * @return return value of {@link CtMethod#getReturnType()}
	 * @throws IllegalArgumentException
	 *             if <code>ctMethod</code> is null
	 * @throws IllegalStateException
	 *             if NotFoundException is thrown
	 */
	public static CtClass getReturnType(CtMethod ctMethod) {
		if (ctMethod == null) throw new IllegalArgumentException("ctMethod is null.");
		try {
			return ctMethod.getReturnType();
		}
		catch (NotFoundException e) {
			throw new IllegalStateException(e.getMessage(), e);
		}
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Checked exception suppressor for {@link CtClass#getSuperclass()}.
	 * </p>
	 * This method converts {@link NotFoundException} to {@link IllegalStateException}.
	 * 
	 * @param ctClass
	 *            target {@link CtClass}
	 * @return return value of {@link CtClass#getSuperclass()}
	 * @throws IllegalArgumentException
	 *             if <code>ctClass</code> is null
	 * @throws IllegalStateException
	 *             if NotFoundException is thrown
	 */
	public static CtClass getSuperClass(CtClass ctClass) {
		if (ctClass == null) throw new IllegalArgumentException("ctClass is null.");
		try {
			return ctClass.getSuperclass();
		}
		catch (NotFoundException e) {
			throw new IllegalStateException(e.getMessage(), e);
		}
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Checked exception suppressor for {@link CtClass#getInterfaces()}.
	 * </p>
	 * This method converts {@link NotFoundException} to {@link IllegalStateException}.
	 * 
	 * @see #getAssignableInterfacesExceptItself(CtClass)
	 * @param ctClass
	 *            target {@link CtClass}
	 * @return return value of {@link CtClass#getInterfaces()}
	 * @throws IllegalArgumentException
	 *             if <code>ctClass</code> is null
	 * @throws IllegalStateException
	 *             if NotFoundException is thrown
	 */
	public static CtClass[] getInterfaces(CtClass ctClass) {
		if (ctClass == null) throw new IllegalArgumentException("ctClass is null.");
		try {
			return ctClass.getInterfaces();
		}
		catch (NotFoundException e) {
			throw new IllegalStateException(e.getMessage(), e);
		}
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Checked exception suppressor for {@link ClassPool#appendClassPath(String)}.
	 * </p>
	 * This method converts {@link NotFoundException} to {@link IllegalStateException}.
	 * 
	 * @param classPool
	 *            target {@link ClassPool}
	 * @param classpath
	 *            target classpath
	 * @return return value of {@link ClassPool#appendClassPath(String)}
	 * @throws IllegalStateException
	 *             if NotFoundException is thrown
	 */
	public static ClassPath appendClassPath(ClassPool classPool, String classpath) {
		if (classPool == null) throw new IllegalArgumentException("classPool is null.");
		if (classpath == null) throw new IllegalArgumentException("classpath is null.");
		try {
			return classPool.appendClassPath(classpath);
		}
		catch (NotFoundException e) {
			// NotFoundException is not thrown is javassist source code in 3.11.0.GA.
			throw new IllegalStateException(e.getMessage(), e);
		}
	}
	
	// Checked(9.10.2009) TODO。コミットしてからここから。
	/**
	 * Checked exception suppressor for {@link CtClass#getDeclaredMethod(String, CtClass[])}.
	 * </p>
	 * This method converts {@link NotFoundException} to {@link IllegalStateException}.
	 * 
	 * @param ctClass
	 *            target {@link CtClass}
	 * @param methodName
	 *            target method name
	 * @param parameterTypes
	 *            target method parameter types
	 * @return return value of {@link CtClass#getDeclaredMethod(String, CtClass[])}
	 * @throws IllegalStateException
	 *             if NotFoundException is thrown
	 */
	public static CtMethod getDeclaredMethod(CtClass ctClass, String methodName, CtClass[] parameterTypes) {
		try {
			return ctClass.getDeclaredMethod(methodName, parameterTypes);
		}
		catch (NotFoundException e) {
			throw new IllegalStateException(e.getMessage(), e);
		}
	}
}
